/* gc-incremental.c
 * The incremental garbage collector.
 *
 * Copyright (c) 1997 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.co.uk>
 */

#define	DBG(s)
#define	FDBG(s)
#define	FTDBG(s)
#define	ADBG(s)

#include "config.h"
#include "config-std.h"
#include "config-mem.h"
#include "gtypes.h"
#include "gc.h"
#include "flags.h"
#include "classMethod.h"
#include "locks.h"
#include "thread.h"
#include "errors.h"
#include "exception.h"
#include "external.h"
#include "lookup.h"
#include "md.h"

static gcList white;
static gcList grey;
static gcList black;
static gcList finalise;

static bool gcRunning = false;

#if defined(GC_INCREMENTAL)
int32 gcAmount;
#endif

/* We run the GC after allocating 1Mbyte of memory.  If we are
 * doing this incrementally, then we will have GC our entire heap
 * by the time we've allocated this much new space.
 */
#define	ALLOCCOUNTGC	(1024*1024)

quickLock gcman;
quickLock finalman;

int gc_mode = GC_DISABLED;	/* GC will be enabled after the first
				 * thread is setup.
				 */

/* Standard GC function sets */
gcFuncs gcNormal =	 { walkConservative, 0 };
gcFuncs gcNoWalk =	 { walkNull, 0 };
gcFuncs gcRoot =	 { walkConservative, (void*)1 };
gcFuncs gcIndirectRoot = { walkIndirectRoot, (void*)1 };

#define	ROOTSIZE	126
typedef struct _rootTable {
	struct _rootTable*	next;
	int			nfree;
	void*			objs[ROOTSIZE];
} rootTable;
static rootTable*	rootObjects;

struct _gcStats gcStats;

static void startGC(void);
static void finishGC(void);

/*
 * Initalise the Garbage Collection system.
 */
void
initGc(void)
{
	gc_heap_initialise();

	RESETLIST(white);
	RESETLIST(grey);
	RESETLIST(black);
	RESETLIST(finalise);

DBG(	printf("white 0x%x\n", &white);
	printf("grey 0x%x\n", &grey);
	printf("black 0x%x\n", &black);
	printf("final 0x%x\n", &finalise);		)
}

/*
 * A reference from one object to another is being created.  Handle
 * the write barrier necessary for incremental collection.
 */
void
soft_addreference(void* f, void* t)
{
	gc_block* finfo;
	gc_block* tinfo;
	int fidx;
	int tidx;

	/* Quit now for null objects */
	if (t == 0) {
		return;
	}

	finfo = GCMEM2BLOCK(f);
	tinfo = GCMEM2BLOCK(t);
	fidx = GCMEM2IDX(finfo, f);
	tidx = GCMEM2IDX(tinfo, t);

	if (GC_GET_COLOUR(finfo, fidx) == GC_COLOUR_BLACK && GC_GET_COLOUR(tinfo, tidx) == GC_COLOUR_WHITE) {
		LOCK();
		REMOVELIST(tinfo);
		APPENDLIST(grey, tinfo);
		GC_SET_COLOUR(tinfo, tidx, GC_COLOUR_GREY);
		UNLOCK();
	}
}

void
markObject(void* mem)
{
	gc_block* info;
	int idx;

	/* If object isn't part of GC, return now */
	if (gc_heap_isobject(mem) == false) {
		return;
	}

	info = GCMEM2BLOCK(mem);
	idx = GCMEM2IDX(info, mem);

	/* If object's been traced before, don't do it again */
	if (GC_GET_COLOUR(info, idx) != GC_COLOUR_WHITE) {
		return;
	}

	/* If we found a new white object, mark it as grey and
	 * move it into the grey list.
	 */
	LOCK();
	GC_SET_COLOUR(info, idx, GC_COLOUR_GREY);
	REMOVELIST(info);
	APPENDLIST(grey, info);
	UNLOCK();
}

void
scanConservative(void* base, size_t size)
{
	void** mem;
	int32 i;

	gcStats.markedmem += size;

	mem = base;
	for (i = (size / sizeof(void*)) - 1; i >= 0; i--) {
		markObject(mem[i]);
	}
}

void
walkConservative(void* base, uint32 size)
{
DBG(	printf("walkConservative: %x-%x\n", base, base+size);
	fflush(stdout);							)

	scanConservative (base, size);
}

/*      
 * Walk an object which cannot reference anything.
 */     
void
walkNull(void* base, uint32 size)
{       
        gcStats.markedmem += size;
}

/*
 * This is a root object which refers to anther object.  A root object
 * has a simple format [void* mem, size_t size].  The 'mem'
 * refers to the object, size it's length
 */
void
walkIndirectRoot(void* base, uint32 size)
{
	rootIndirect* mem;

	mem = (rootIndirect*)base;
	(*mem->funcs->walk)(mem->mem, mem->len);
}

/*
 * Walk a GC'able object.  Mark the object as black then for each pointer
 * in the object which refers to another GC object, grey the object
 * and insert into the grey list.
 */
void
walkBlock(gc_block* info)
{
	int idx;

	/* Move object to black list.  We do this before we start walking
	 * the block incase we reference another object in the same block
	 * and so move this object back onto the grey list.
	 */
	REMOVELIST(info);
	APPENDLIST(black, info);

	for (idx = 0; idx < info->nr; idx++) {
		if (GC_GET_COLOUR(info, idx) == GC_COLOUR_GREY) {
			GC_SET_COLOUR(info, idx, GC_COLOUR_BLACK);
			gcStats.markedobj += 1;
			(*GC_GET_FUNCS(info, idx)->walk)(GCBLOCK2MEM(info, idx), GCBLOCKSIZE(info));
		}
	}
}

/*
 * Walk a bit of memory.
 */
void
walkMemory(void* mem)
{
	gc_block* info;
	int idx;

	info = GCMEM2BLOCK(mem);
	idx = GCMEM2IDX(info, mem);

	REMOVELIST(info);
	APPENDLIST(black, info);
	GC_SET_COLOUR(info, idx, GC_COLOUR_BLACK);

	gcStats.markedobj += 1;
	(*GC_GET_FUNCS(info, idx)->walk)(mem, GCBLOCKSIZE(info));
}

/*
 * The Garbage Collector sits in a loop starting a collection, waiting
 * until it's finished incrementally, then tidying up before starting
 * another one.
 */
void
gcMan(void)
{
	gc_block* info;
	int idx;

	/* All threads start with interrupts disabled */
	intsRestore();

	lockMutex(&gcman);

	for(;;) {
#if defined(GC_INCREMENTAL)
		startGC();

		gcRunning = false;
		waitCond(&gcman, 0);
		assert(gcRunning == true);

		LOCK();

		/* Walk the threads.  Although we do this initially it is
		 * possible that there's an object on the thread stack which
		 * isn't referenced anywhere else but is live.
		 */
		walkLiveThreads();

		/* Walk any remaining grey objects - shouldn't be necessary
		 * unless GC has been activated explicitly.
		 */
		for (info = grey.cnext; info != &grey; info = grey.cnext) {
			walkBlock(info);
		}

		/* Now walk any white objects which will be finalied.  They
		 * may get reattached, so anything they reference must also
		 * be live just in case.
		 */
		for (info = white.cnext; info != &white; info = info->cnext) {
			for (idx = 0; idx < info->nr; idx++) {
				if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
					gcStats.markedobj += 1;
					(*GC_GET_FUNCS(info, idx)->walk)(GCBLOCK2MEM(info, idx), GCBLOCKSIZE(info));
				}
			}
		}

		/* We may now have more grey objects, so walk them */
		for (obj = grey.cnext; obj != &grey; obj = grey.cnext) {
			walkBlock(obj);
		}

		finishGC();

		UNLOCK();
#else
		gcRunning = false;
		waitCond(&gcman, 0);
		assert(gcRunning == true);

		LOCK();

		startGC();

		for (info = grey.cnext; info != &grey; info = grey.cnext) {
			walkBlock(info);
		}

		/* Now walk any white objects which will be finalied.  They
		 * may get reattached, so anything they reference must also
		 * be live just in case.
		 */
		for (info = white.cnext; info != &white; info = info->cnext) {
			for (idx = 0; idx < info->nr; idx++) {
				if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
					gcStats.markedobj += 1;
					(*GC_GET_FUNCS(info, idx)->walk)(GCBLOCK2MEM(info, idx), GCBLOCKSIZE(info));
				}
			}
		}

		/* We may now have more grey objects, so walk them */
		for (info = grey.cnext; info != &grey; info = grey.cnext) {
			walkBlock(info);
		}

		finishGC();

		UNLOCK();

#endif

		if (flag_gc) {
			fprintf(stderr, "<GC: total %dK, alloc %dK, marked %dK, freeing %dK>\n", gcStats.totalmem/1024, gcStats.allocmem/1024, gcStats.markedmem/1024, (gcStats.totalmem > gcStats.markedmem ? (gcStats.totalmem - gcStats.markedmem)/1024 : 0));
		}
		gcStats.totalmem -= gcStats.freedmem;
		gcStats.totalobj -= gcStats.freedobj;
		gcStats.allocobj = 0;
		gcStats.allocmem = 0;

		/* Wake up anyone waiting for the GC to finish */
		broadcastCond(&gcman);
	}
}

/*
 * Start the GC process by scanning the root and thread stack objects.
 */
static
void
startGC(void)
{
	void* obj;
	gc_block* info;
	int idx;
	int i;
	rootTable* tptr;

	gcStats.freedmem = 0;
	gcStats.freedobj = 0;
	gcStats.markedobj = 0;
	gcStats.markedmem = 0;

	/* Walk the root objects */
	for (tptr = rootObjects; tptr != 0; tptr = tptr->next) {
		for (i = 0; i < tptr->nfree; i++) {
			obj = tptr->objs[i];
			info = GCMEM2BLOCK(obj);
			idx = GCMEM2IDX(info, obj);
			GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
			markObject(obj);
		}
	}

	/* Walk the thread objects */
	walkLiveThreads();
}

/*
 * Finish off the GC process.  Any unreached (white) objects are moved
 * for finalising and the finaliser woken.
 * The reached (black) objects are moved onto the now empty white list
 * and recoloured white.
 */
static
void
finishGC(void)
{
	gc_block* info;
	int idx;

	/* There shouldn't be any grey objects at this point */
	assert(grey.cnext == &grey);

	/* Any white objects should now be finalised */
	if (white.cnext != &white) {
		do {
			info = white.cnext;
			REMOVELIST(info);

			/* Walk block looking to see if any objects need
			 * finalizing.  If so, attach the whole block to the
			 * finalizer queue.
			 */
			for (idx = 0; idx < info->nr; idx++) {
				if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
					APPENDLIST(finalise, info);
					goto donext;
				}
			}

			/* No finalizers, so just free all the white objects */
			for (idx = 0; idx < info->nr; idx++) {
				if (GC_GET_COLOUR(info, idx) == GC_COLOUR_WHITE) {
FDBG(					printf("freeObject %p size %d\n",
						info, info->size);
						fflush(stdout);		)
					gcStats.freedmem += GCBLOCKSIZE(info);
					gcStats.freedobj += 1;
					gc_heap_free(GCBLOCK2MEM(info, idx));
				}
			}
			donext:;

		} while (white.cnext != &white);

		lockMutex(&finalman);
		signalCond(&finalman);
		unlockMutex(&finalman);
	}

FTDBG(	printf("Freed %d objects of %dK\n", gcStats.freedobj,
		gcStats.freedmem/1024);					)

	/* Now move the black objects back to the white queue for next time.
	 */
	while (black.cnext != &black) {
		info = black.cnext;
		REMOVELIST(info);
		APPENDLIST(white, info);
		for (idx = 0; idx < info->nr; idx++) {
			if (GC_GET_COLOUR(info, idx) == GC_COLOUR_BLACK) {
				GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
			}
		}
	}
}

/*
 * The finaliser sits in a loop waiting to finalise objects.  When a
 * new finalised list is available, it is woken by the GC and finalises
 * the objects in turn.  An object is only finalised once after which
 * it is deleted.
 */
void
finaliserMan(void)
{
	gc_block* info;
	int idx;

	/* All threads start with interrupts disabled */
	intsRestore();

	lockMutex(&finalman);

	for (;;) {
		waitCond(&finalman, 0);
		while (finalise.cnext != &finalise) {
			LOCK();
			info = finalise.cnext;
			REMOVELIST(info);
			APPENDLIST(white, info);
			UNLOCK();

			for (idx = 0; idx < info->nr; idx++) {
				if (GC_GET_STATE(info, idx) == GC_STATE_NEEDFINALIZE) {
					/* Objects are only finalised once */
					GC_SET_STATE(info, idx, GC_STATE_FINALIZED);
					GC_SET_COLOUR(info, idx, GC_COLOUR_WHITE);
					/* Call finaliser */
					(*GC_GET_FUNCS(info, idx)->final)(GCBLOCK2MEM(info, idx));
				}
			}
		}
	}
}

/*
 * Explicity invoke the garbage collecto and wait for it to complete.
 */
void
invokeGC(void)
{
	lockMutex(&gcman);
	if (gcRunning == false) {
		gcRunning = true;
		signalCond(&gcman);
	}
	waitCond(&gcman, 0);
	unlockMutex(&gcman);
}

/*
 * Allocate a new object.  The object is attached to the white queue.
 * After allocation, if incremental collection is active we peform
 * a little garbage collection.  If we finish it, we wakeup the garbage
 * collector.
 */
void*
gc_malloc(size_t size, gcFuncs* funcs)
{
	gc_block* info;
	rootTable* t;
	void* mem;
	int i;

	LOCK();

	mem = gc_heap_malloc(size);
	if (mem == 0) {
		throwOutOfMemory();  
	}
ADBG(	printf("gc_malloc: 0x%x (%d)\n", mem, size);			)

	info = GCMEM2BLOCK(mem);
	i = GCMEM2IDX(info, mem);

	gcStats.totalmem += GCBLOCKSIZE(info);
	gcStats.totalobj += 1;
	gcStats.allocmem += GCBLOCKSIZE(info);
	gcStats.allocobj += 1;

	GC_SET_FUNCS(info, i, funcs);

	/* Determine whether we need to finalise or not */
	if (funcs->final == 0 || funcs->final == (void*)1) {
		GC_SET_STATE(info, i, GC_STATE_NORMAL);
	}
	else {
		GC_SET_STATE(info, i, GC_STATE_NEEDFINALIZE);
	}

	/* If object is a root object, attach it to root list */
	if (funcs->final == (void*)1) {
		t = rootObjects;
		if (t == 0 || t->nfree == ROOTSIZE) {
			rootObjects = checked_calloc(sizeof(rootTable), 1);
			rootObjects->next = t;
			t = rootObjects;
		}
		t->objs[t->nfree] = mem;
		t->nfree++;
	}

#if defined(GC_INCREMENTAL)
	GC_SET_COLOUR(info, i, GC_COLOUR_GREY);
	if (ONLIST(info)) {
		REMOVELIST(info);
	}
	APPENDLIST(grey, info);
#else
	GC_SET_COLOUR(info, i, GC_COLOUR_WHITE);
	if (ONLIST(info)) {
		REMOVELIST(info);
	}
	APPENDLIST(white, info);
#endif

#if defined(GC_INCREMENTAL)
	/* Perform a little GC as we go */
	if (grey.cnext != &grey) {
		gcAmount += (gcStats.totalmem * GC_OBJECT_SIZE(obj)) / ALLOCCOUNTGC;
		while (gcAmount > 0) {
			gcAmount -= GC_OBJECT_SIZE(grey.cnext);
			walkBlock(grey.cnext);
			if (grey.cnext == &grey) {
				UNLOCK();
				invokeGC();
				goto out;
			}
		}
		UNLOCK();
		out:;
	}
	else {
		UNLOCK();
	}
#else
	UNLOCK();
#endif
	return (mem);
}

/*
 * Attach a random piece of memory to the GC system.  The object is made
 * a root, which is generally what is wanted.
 */
void
gc_attach(void* mem, size_t len, gcFuncs* funcs)
{
	rootIndirect* root;

	root = gc_malloc(sizeof(rootIndirect), &gcIndirectRoot);
	root->mem = mem;
	root->len = len;
	root->funcs = funcs;
}

/*
 * Set a specific finaliser on some memory.
 */
void
gc_set_finalizer(void* mem, gcFuncs* funcs)
{
	gc_block* info;
	int idx;

	info = GCMEM2BLOCK(mem);
	idx = GCMEM2IDX(info, mem);

	GC_SET_STATE(info, idx, GC_STATE_NEEDFINALIZE);
	GC_SET_FUNCS(info, idx, funcs);
}
